/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <wchar.h>

#include <iomanip>
#include <iostream>

#include <stdlib.h>
#include <getopt.h>

#include "libtkcommon/config.h"
#include "libtkcommon/tkintl.h"
#include "libtkanalysis/analyser.h"
#include "libtkdb/database.h"
#ifdef HAVE_GNUPLOT
#include "libtkanalysis/gnuplot.h"
#endif /* HAVE_GNUPLOT */

using namespace Tk;
using namespace Analysis;

void
print_help ()
{
  std::cout  << PACKAGE_NAME << " version " << PACKAGE_VERSION << std::endl;
  std::wcout << L"Copyright (C) 2005 Robert Ham (rah@bash.sh)" << std::endl;
  std::wcout << std::endl;
  std::wcout << L"This program comes with ABSOLUTELY NO WARRANTY.  You are licensed to use it" << std::endl;
  std::wcout << L"under the terms of the GNU General Public License, version 3 or later.  See" << std::endl;
  std::wcout << L"the COPYING file that came with this software for details." << std::endl;
  std::wcout << std::endl;
  std::wcout << L" -h, --help                  " << _("Display this help info") << std::endl;
  std::wcout << L" -n, --no-plot               " << _("Don't display gnuplots") << std::endl;
  std::wcout << L" -d, --dsn <host>            " << _("Connect to data source <dsn>") << std::endl;
  std::wcout << L" -u, --username <user>       " << _("Connect as <user>") << std::endl;
  std::wcout << L" -p, --password <pass>       " << _("Connect with <pass>") << std::endl;
  std::wcout << L" -q, --query <file>          " << _("Analyse experiments matching the query in <file>") << std::endl;
  std::wcout << std::endl;
}

int
main (int argc, char ** argv)
{
  String data_source_name("TkResolver++");
  String username("psi");
  String password("isp");
  String query_file;
  wchar_t * wstr;
  bool plot(true);

  int opt;
  const char * options = "hnd:u:p:q:";
  struct option long_options[] = {
    { "help",        no_argument,       NULL, 'h' },
    { "no-plot",     no_argument,       NULL, 'n' },
    { "dsn",         required_argument, NULL, 'd' },
    { "username",    required_argument, NULL, 'u' },
    { "password",    required_argument, NULL, 'p' },
    { "query",       required_argument, NULL, 'q' },
    { 0, 0, 0, 0 }
  };
  
  while ((opt = getopt_long (argc, argv, options, long_options, NULL)) != -1) {
    switch (opt) {
      
    case 'd':
      data_source_name = optarg;
      break;
      
    case 'u':
      username = optarg;
      break;
      
    case 'p':
      password = optarg;
      break;
      
    case 'q':
      query_file = optarg;
      break;
      
    case 'n':
      plot = false;
      break;
      
    case 'h':
      print_help ();
      exit (0);
      break;
    case ':':
    case '?':
      print_help ();
      exit (1);
      break;
    }
  }
  
  if (query_file.size() < 1)
    {
      print_help();
      std::wcerr << _(L"Query file must be specified") << std::endl;
      exit(1);
    }
  
  Glib::thread_init();
  
  try
    {
      Db::Database database(data_source_name, username, password);
	  
      std::wcout << _(L"Analysing results...")
		 << std::flush;
      
      //      std::wcout << L"default precision was: " << mpf_get_default_prec() << std::endl;
      mpf_set_default_prec(1024);
      Analyser analyser(database, query_file);
	  
      std::wcout << _(L"done")
		 << std::endl
		 << std::endl;

      
      for (TraceStatisticsList::iterator trace_iter(analyser._statistics.begin());
	   trace_iter != analyser._statistics.end();
	   ++trace_iter)
	{
	  std::wcout << L"=======  Trace count "
		     << String(trace_iter->_trace_count)
		     << L"  ======="
		     << std::endl
	    ;
	  
	  AlgorithmStatisticsList &algo_stats( trace_iter->_algorithm_stats );
	  
	  for (AlgorithmStatisticsList::iterator algo_iter( algo_stats.begin() );
	       algo_iter != algo_stats.end();
	       ++algo_iter)
	    {
	      std::wcout << L" =====  Algorithm '"
			 << String(algo_iter->_algorithm_spec)
			 << L"' ====="
			 << std::endl
		;

	      StatisticsList &intent_stats( algo_iter->_intent_stats );
	      
	      for (StatisticsList::iterator intent_iter( intent_stats.begin() );
		   intent_iter != intent_stats.end();
		   ++intent_iter)
		{
		  Statistics &stats( *intent_iter );

		  mp_exp_t exponent(100);
		  int base(10);
		  size_t digits(20);
		  std::wcout << L"  === Intent code '"
			     << String(stats._intent_code)
			     << L"' ==="
			     << std::endl
		    
			     << L"   "
			     << _(L"n:            ")
			     << String(stats._n)
			     << std::endl
		    
			     << L"   "
			     << _(L"experiments:  ")
			     << String(stats._experiment_count)
			     << std::endl
		    
			     << L"   "
			     << _(L"pos min:      ")
			     << String(stats._position_min)
			     << std::endl
		    
			     << "   "
			     << _(L"pos max:      ")
			     << String(stats._position_max)
			     << std::endl
		    
			     << "   "
			     << _(L"exp pos mean: ")
			     << String(stats._experiment_position_mean)
			     << std::endl
		    
			     << "   "
			     << _(L"exp pos var:  ")
			     << String(stats._experiment_position_variance)
			     << std::endl
		    ;
		  
		  std::wcout << "   "
			     << _(L"Offsets:      ")
			     << std::endl;
		  for (OffsetStatisticsMap::iterator offset_iter( stats._offsets._offset_stats.begin() );
		       offset_iter != stats._offsets._offset_stats.end();
		       ++offset_iter)
		    {
		      std::wcout << L"      "
				 << offset_iter->first
				 << L":"
				 << std::endl
			
				 << L"         "
				 << L"theor. p: "
				 << String( algo_iter->_offset_probabilities[offset_iter->first] )
				 << std::endl
			
				 << L"         "
				 << L"mean:     "
				 << String( offset_iter->second._mean )
				 << std::endl
			
				 << L"         "
				 << L"var:      "
				 << String( offset_iter->second._variance )
				 << std::endl
			
				 << L"         "
				 << L"skew:     "
				 << String( offset_iter->second._skewness )
				 << std::endl
			
				 << L"         "
				 << L"rss:      "
				 << String( offset_iter->second._rss )
				 << std::endl
			;
		    }
	  
#ifdef HAVE_GNUPLOT
		  if (plot)
		    Gnuplot plot(stats._intent_code, stats._offsets);
#endif /* HAVE_GNUPLOT */
		  
		  StatisticsList::iterator inner_iter(intent_iter);
		  for (++inner_iter;
		       inner_iter != intent_stats.end();
		       ++inner_iter)
		    {
		      Statistics &inner_stats(*inner_iter);
		      
#ifdef HAVE_GNUPLOT
		      if (plot)
			Gnuplot plot(stats._intent_code, stats._offsets,
				     inner_stats._intent_code, inner_stats._offsets);
#endif /* HAVE_GNUPLOT */
		      
		      if (stats._experiment_count > 1
			  && inner_stats._experiment_count > 1)
			{
			  TTest t(stats, inner_stats);
			  std::wcout << L"   "
				     << stats._intent_code + L"-" + inner_stats._intent_code
				     << L" t-test:"
				     << std::endl
			    
				     << L"      "
				     << _(L"t score: ")
				     << String(t._t)
				     << std::endl
			    
				     << L"      "
				     << _(L"df:      ")
				     << String(t._df)
				     << std::endl
			    
				     << L"      "
				     << _(L"p:       ")
				     << String(t._p)
				     << std::endl
			    
				     << L"      "
				     << _(L"α:       ")
				     << String(static_cast<Real>(100 *(1 - t._p)), 3)
				     << L"%"
				     << std::endl
			    ;
			}
		    }
		}
	    }
	  std::wcout << std::endl;
	  
	}
    } 
   catch (const NotEnoughResults & ner)
     {
       std::wcout << _(L"not enough results")
		  << std::endl;
       exit(0);
     }

  exit (0);
}
