/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2006 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "gnuplot.h"

using namespace Tk::Analysis;


Gnuplot::Arguments
Gnuplot:: arguments(unsigned set_count, unsigned set_size)
{
  Arguments arguments;
  Argument argument;
  int i, s;
  unsigned long long line_no;

  arguments.push_back("gnuplot");

  line_no = 1;
  for (i = 0; i < set_count; ++i)
    {
      /*
       * Offset counts
       */
      arguments.push_back("-xrm");
      
      argument = "gnuplot*line";
      argument += Tk::String(line_no);
      argument += "Color: ";
      
      switch (i)
	{
	case 0: argument += "Black"; break;
	case 1: argument += "DimGray"; break;
	case 2: argument += "Gray"; break;
	}
      arguments.push_back(argument);

      ++line_no;
      
      /*
       * Probabilities
       */
      for (s = 0; s < set_size; ++s)
	{
	  /* line colour */
	  arguments.push_back("-xrm");
	  
	  argument = "gnuplot*line";
	  argument += String(line_no);
	  argument += "Color: ";
	  
	  switch (i)
	    {
	      case 0:
		switch (s)
		  {
		  case 0: argument += "Red"; break;
		  case 1: argument += "FireBrick"; break;
		  case 2: argument += "Brown"; break;
		  case 3: argument += "IndianRed"; break;
		  case 4: argument += "RosyBrown"; break;
		  }
		break;
	    case 1:
		switch (s)
		  {
		  case 0: argument += "Navy"; break;
		  case 1: argument += "RoyalBlue"; break;
		  case 2: argument += "CornflowerBlue"; break;
		  case 3: argument += "LightSteelBlue"; break;
		  case 4: argument += "LightSteelGray"; break;
		  }
	    }
	  
	  arguments.push_back(argument);

	  /* line dashes */
	  arguments.push_back("-xrm");
	  
	  argument = "gnuplot*line";
	  argument += String(line_no);
	  argument += "Dashes: ";
	  
	  switch (s)
	    {
	    case 0: argument += "0"; break;
	    case 1: argument += "31"; break;
	    case 2: argument += "4111"; break;
	    }
	  
	  arguments.push_back(argument);
	  
	  ++line_no;
	}
    }
  
  return arguments;
}

void
Gnuplot:: spawn(const Arguments &arguments)
{
  int gnuplot_fd;
  
  Glib::spawn_async_with_pipes("/",
			       arguments,
			       Glib::SPAWN_SEARCH_PATH,
			       sigc::slot< void >(),
			       0,
			       &gnuplot_fd);
  
  _gnuplot_file = new FileStream(gnuplot_fd, std::ios_base::out);
  _gnuplot = new OutputStream(_gnuplot_file);
  
  *_gnuplot << L"set terminal x11 persist" << std::endl;
  *_gnuplot << L"set xlabel \"Normalised screen position\"" << std::endl;
  *_gnuplot << L"set ylabel \"P(Offset)\"" << std::endl;
  *_gnuplot << L"set y2label \"Offset count\"" << std::endl;
  *_gnuplot << L"set ytics nomirror" << std::endl;
  *_gnuplot << L"set y2tics autofreq" << std::endl;
  
  *_gnuplot << L"plot";
}

void
Gnuplot:: plot(const String &name,
	       const OffsetList &offsets,
	       bool first,
	       bool last)
{
  String prefix;
  if (name.size() > 0)
    {
      prefix = (name == "NULL") ? L"CT" : name;
      prefix += L" - ";
    }
  
  /* Offset counts */
  *_gnuplot << (first ? L"" : L",")
	    << L" '-' title \""
	    << prefix
	    << L"Offset counts\" axes x1y2 with lines";
	
  for (OffsetList::const_iterator iter(offsets.begin());
       iter != offsets.end();
       ++iter)
    *_gnuplot << L", '-' title \""
	      << prefix
	      << *iter
	      << L"\" smooth acsplines";
  if (last)
    *_gnuplot << std::endl;
}
  

void
Gnuplot:: position_counts(const PositionCounts &position_counts)
{
  for (PositionCounts::const_iterator iter( position_counts.begin() );
       iter != position_counts.end();
       ++iter)
    {
      *_gnuplot << String( static_cast<Real>(iter->first) )
		<< L" "
		<< String( iter->second )
		<< std::endl;
    } 
  *_gnuplot << L"e" << std::endl;
}

void
Gnuplot:: position_function(const PositionCounts &position_counts,
			    const PositionFunction &positions)
{
  
  for (PositionFunction::const_iterator pos_iter( positions.begin() );
       pos_iter != positions.end();
       ++pos_iter)
    {
      //      if (pos_iter->second._offset_distribution.size() == offset_count)
	*_gnuplot << String( static_cast<Real>(pos_iter->first) )
		  << L" "
		  << String( static_cast<Real>(pos_iter->second) )
		  << L" "
		  << String( static_cast<Real>( position_counts.find(pos_iter->first)->second * static_cast<Rational>(1000000) ) )
		  << std::endl;
    }
  *_gnuplot << L"e" << std::endl;
}

void
Gnuplot:: offsets(const OffsetStatisticsSet &stats)
{
  for (OffsetStatisticsMap::const_iterator iter( stats._offset_stats.begin() );
       iter != stats._offset_stats.end();
       ++iter)
    position_function(stats._position_totals,
		      iter->second._position_function);
}

void
Gnuplot:: statistics(const OffsetStatisticsSet& stats)
{
  position_counts(stats._position_totals);
  offsets(stats);
}

Gnuplot:: Gnuplot(const String &name, const OffsetStatisticsSet &stats)
{
  spawn( arguments(1, stats._offsets.size()) );
  plot(name, stats._offsets, true, true);
  statistics(stats);
}

Gnuplot:: Gnuplot(const String &name_a,
		  const OffsetStatisticsSet &stats_a,
		  const String &name_b,
		  const OffsetStatisticsSet &stats_b)
{
  spawn( arguments(2, stats_a._offsets.size()) );
  plot(name_a, stats_a._offsets, true, false);
  plot(name_b, stats_b._offsets, false, true);
  statistics(stats_a);
  statistics(stats_b);
}

Gnuplot:: ~Gnuplot()
{
  _gnuplot_file->close();
  delete _gnuplot;
  delete _gnuplot_file;
}
