/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "libtkcommon/tkintl.h"
#include "libtkcommon/debug.h"

#include "libtkquery/expression.h"
#include "libtkquery/identifier.h"
#include "libtkquery/parse_error.h"
#include "libtkdb/exception.h"

using namespace Tk::Query;

static const Tk::String obrkt("(");
static const Tk::String cbrkt(")");
static const Tk::String space(" ");

Boolean:: Boolean(bool b)
  : _val(b)
{
}

Query *
Boolean:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Boolean" << std::endl;
  Query *query( new Query(database) );
  
  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}

Integer:: Integer(const char *text)
  : _val(atoi(text))
{
}

Query *
Integer:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Integer (" << _val << ")" << std::endl;
  Query *query( new Query(database) );

  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}


Real:: Real(const char *text)
  : _val(strtod(text, NULL))
{
}

Query *
Real:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Real" << std::endl;
  Query *query( new Query(database) );

  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}


String:: String(const Tk::String &str)
{
  // trim quote marks
  Tk::String::const_iterator c( str.begin() );
  for (++c; c != str.end();)
    {
      Tk::String::value_type chr( *c );
      
      if (++c != str.end())
	_val.push_back(chr);
    }
}
	
Query *
String:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building String" << std::endl;
  Query *query( new Query(database) );

  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}


Name:: Name(const Tk::String &name)
{
  // trim start and end marks
  Tk::String::const_iterator c( name.begin() );
  for (++c; c != name.end();)
    {
      Tk::String::value_type chr( *c );
      
      if (++c != name.end())
	_name.push_back(chr);
    }
}

const Tk::String &
Name:: get_name() const
{
  return _name;
}


Alpha:: Alpha (const char *text)
  : _text(text)
{
}

const Tk::String &
Alpha:: get_text() const
{
  return _text;
}


Identifier:: Identifier(Alpha *alpha1,
			const Location &alpha1_loc,
			Alpha *alpha2,
			const Location &alpha2_loc)
  : _domain( resolve_domain(alpha1_loc, alpha1->get_text()) ),
    _domain_loc( new Location(alpha1_loc) ),
    _property_name( alpha2->get_text() ),
    _property_loc( alpha2_loc )
{
}

Identifier:: Identifier(Alpha *alpha,
			const Location &alpha_loc,
			Name *name,
			const Location &name_loc)
  : _domain( resolve_domain(alpha_loc, alpha->get_text()) ),
    _domain_loc( new Location(alpha_loc) ),
    _property_name( name->get_name() ),
    _property_loc(name_loc)
{
}

Identifier:: Identifier(Name *name, const Location &name_loc)
  : _domain_loc(0),
    _property_name( name->get_name() ),
    _property_loc(name_loc)
{
}

Identifier:: Identifier(Alpha *alpha, const Location &alpha_loc)
  : _domain_loc(0),
    _property_name( alpha->get_text() ),
    _property_loc(alpha_loc)
{
}

Identifier:: ~Identifier()
{
  if (_domain_loc)
    delete _domain_loc;
}

Tk::String
Identifier::resolve_domain(const Location &loc,
			   const Tk::String &partial)
{
  Tk::Strings domains;
  domains.push_back("experiment");
  domains.push_back("operator");
  domains.push_back("location");
  
  for (Tk::Strings::iterator iter( domains.begin() );
       iter != domains.end();
       )
    {
      if (iter->size() < partial.size())
	iter = domains.erase(iter);
      else
	++iter;
    }
  
  for (unsigned i(0); i < partial.size(); ++i)
    {
      for (Tk::Strings::iterator iter( domains.begin() );
	   iter != domains.end();
	   )
	{
	  if ((*iter)[i] != partial[i])
	    iter = domains.erase(iter);
	  else
	    ++iter;
	}
      
      if (domains.size() < 2)
	break;
    }
  
  if (domains.size() != 1)
    throw SemanticError(loc,
			std::string(_("Ambiguous domain specifier `"))
			+ static_cast<std::string>(partial)
			+ _("'")
			);
  
  return *domains.begin();
}

Query *
Identifier:: build_query(Tk::Db::Database *database, const Tk::String &query_domain) const
{
  Tk::String id_domain;
  if (_domain.size() > 0)
    id_domain = _domain;
  else
    id_domain = query_domain;
  
  //WDEBUG_OUT << L"building Identifier (name: '" << _name << "')" << std::endl;
  Query *query( new Query(database) );
  
  TableAlias property( query->add_table("property") );
  TableAlias property_type( query->add_table("property_type") );
  TableAlias domain_property( query->add_table(id_domain + Tk::String("_property")) );
  
  TableAlias query_domain_alias( Query::get_table_abbrev(query_domain) );
  Tk::String query_domain_id( query_domain + Tk::String("_id") );
  
  const Tk::String property_id("property_id");
  const Tk::String property_type_id("property_type_id");
  
  if ("experiment" == query_domain)
    {
      if ("operator" == id_domain)
	{
	  const Tk::String operator_id("operator_id");
	  
	  TableAlias experiment_op( query->add_table("experiment_operator") );
	  TableAlias op( query->add_table("operator") );
	  
	  query->add_link_clause(experiment_op, query_domain_id, query_domain_alias, query_domain_id);
	  query->add_link_clause(op, operator_id, experiment_op, operator_id);
	  query->add_link_clause(domain_property, operator_id, op, operator_id);
	}
      else if ("location" == id_domain)
	{
	  const Tk::String location_id("location_id");
	  
	  TableAlias location( query->add_table("location") );
	  
	  query->add_link_clause(location, location_id, query_domain_alias, location_id);
	  query->add_link_clause(domain_property, location_id, location, location_id);
	}
      else
	{
	  query->add_link_clause(domain_property, query_domain_id, query_domain_alias, query_domain_id);
	}
    }
  else
    {
      if (id_domain != query_domain)
	throw SemanticError(*_domain_loc,
			    std::string("Invalid identifier domain `")
			    + static_cast<std::string>(id_domain)
			    + "' with query domain `"
			    + static_cast<std::string>(query_domain)
			    + "'"
			    );
      
      query->add_link_clause(domain_property, query_domain_id, query_domain_alias, query_domain_id);
    }
  
  TableAlias property_value;
  try
    {
      property_value = query->add_property_value_table(id_domain, _property_name);
    }
  catch (Tk::Db::UnknownType &ut)
    {
      throw SemanticError(_property_loc,
			  static_cast<std::string>("Unknown property `")
			  + static_cast<std::string>(_property_name)
			  + "' in domain `"
			  + static_cast<std::string>(id_domain)
			  + "'");
    }
  
  query->add_link_clause(property, property_id, domain_property, property_id);
  query->add_link_clause(property_value, property_id, property, property_id);
  query->add_link_clause(property_type, property_type_id, property, property_type_id);
  
  Tk::Db::Data::Values domain_params;
  domain_params.push_back(Tk::Db::Data::Value(Tk::Db::Data::String, static_cast<Tk::Db::Data::Datum>(id_domain)));
  query->add_link_clause(property_type + Tk::String(".domain = ?"), domain_params);
  
  Tk::Db::Data::Values name_params;
  name_params.push_back(Tk::Db::Data::Value(Tk::Db::Data::String, static_cast<Tk::Db::Data::Datum>(_property_name)));
  query->add_link_clause(property_type + Tk::String(".name = ?"), name_params);
  
  query->set_expr(property_value + Tk::String(".property_value"));
  
  return query;
}


Op:: Op(const char *text)
  : _op(text)
{}


Binary:: Binary(Value *val1, Op *op, Value *val2)
  : _val1(val1),
    _op(op),
    _val2(val2)
{}

Binary:: ~Binary()
{
  delete _val1;
  delete _op;
  delete _val2;
}

Query *
Binary:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Binary" << std::endl;
  QueryPtr val1_query( _val1->build_query(database, domain) );
  QueryPtr val2_query( _val2->build_query(database, domain) );
  
  Query *query( new Query(database) );
  
  query->append_tables(*val1_query);
  query->append_tables(*val2_query);
  
  /* sandwich values and parameters */
  Tk::String expr( obrkt + val1_query->get_expr() + space + _op->get_op() + space + val2_query->get_expr() + cbrkt );
  Tk::Db::Data::Values expr_params( val1_query->get_expr_params() );
  expr_params.append(val2_query->get_expr_params());
  
  query->set_expr(expr, expr_params);
  
  return query;
}


Unary:: Unary (Op *op, Value *val)
  : _op(op),
    _val(val)
{}

Unary:: ~Unary()
{
  delete _op;
  delete _val;
}

Query *
Unary:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Unary" << std::endl;
  QueryPtr val_query( _val->build_query(database, domain) );
  
  Query *query( new Query(database) );
  query->append_tables(*val_query);

  Tk::Db::Data::Values params( val_query->get_expr_params() );
  params.push_back(false);
  
  query->set_expr(obrkt + val_query->get_expr() + Tk::String(" = ?") + cbrkt, params);
  
  return query;
}


Logical:: Logical (Expression *expr1, Op *lop, Expression *expr2)
  : _expr1(expr1),
    _lop(lop),
    _expr2(expr2)
{}

Logical:: ~Logical()
{
  delete _expr1;
  delete _lop;
  delete _expr2;
}

Query *
Logical:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Logical" << std::endl;
  Tk::String op;
  if (_lop->get_op() == "&")
    op = "and";
  else
    op = "or";
  
  QueryPtr expr1_query( _expr1->build_query(database, domain) );
  QueryPtr expr2_query( _expr2->build_query(database, domain) );
  
  Query *query( new Query(database) );
  
  query->append_tables(*expr1_query);
  query->append_tables(*expr2_query);
  
  /* sandwich values and parameters */
  Tk::String expr( expr1_query->get_expr() + space + op + space + expr2_query->get_expr() );
  Tk::Db::Data::Values expr_params( expr1_query->get_expr_params() );
  expr_params.append(expr2_query->get_expr_params());
  
  query->set_expr(expr, expr_params);
  
  return query;

  //  return _expr1->build_query(components) + space + op + space + _expr2->build_query(components);
}


Bracketed:: Bracketed (Expression *expr)
  : _expr(expr)
{}

Bracketed:: ~Bracketed()
{
  delete _expr;
}

Query *
Bracketed:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Bracketed" << std::endl;
  Query *query( _expr->build_query(database, domain) );
  query->set_expr( obrkt + query->get_expr() + cbrkt, query->get_expr_params() );
  return query;
}
