/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __TK_EXPRESSION_H__
#define __TK_EXPRESSION_H__

#include "libtkcommon/tkstring.h"
#include "libtkcommon/numerical-types.h"
#include "libtkcommon/debug.h"

#include "libtkquery/query.h"

namespace Tk {
  namespace Query {
    
    class Query;
    
    class QueryBuilder
      {
      public:
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const = 0;
      };
    
    
    class Value : public QueryBuilder
      {
      public:
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const = 0;
      };
    
    class Boolean : public Value
      {
      private:
	bool _val;
      public:
	Boolean (bool b);
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };
    
    class Integer : public Value
      {
      private:
	long _val;
      public:
	Integer (const char *text);
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };
    
    class Real : public Value
      {
      private:
	double _val;
      public:
	Real (const char *text);
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };
    
    class String : public Value
      {
      private:
	Tk::String _val;
      public:
	String (const Tk::String &str);
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };
    
    class Name
      {
      private:
	Tk::String _name;
      public:
	Name (const Tk::String &name);
	const Tk::String &get_name() const;
      };
    
    class Alpha
      {
      private:
	Tk::String _text;
      public:
	Alpha (const char *text);
	const Tk::String &get_text() const;
      };
    
    class Op
      {
      private:
	Tk::String _op;
      public:
	Op (const char *text);
	Tk::String get_op() const { return _op; }
      };
    
    
    class Expression : public QueryBuilder
      {
      public:
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const = 0;
      };
    typedef std::auto_ptr<Expression> ExpressionPtr;
    
    class Clause : public Expression
      {
      public:
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const = 0;
      };
    
    class Binary : public Clause
      {
      private:
	Value *_val1;
	Op *_op;
	Value *_val2;
      public:
	Binary (Value *val1, Op *op, Value *val2);
	~Binary();
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };

    class Unary : public Clause
      {
      private:
	Op *_op;
	Value *_val;
      public:
	Unary (Op *op, Value *val);
	~Unary();
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };
    
    
    class Logical : public Expression
      {
      private:
	Expression *_expr1;
	Op *_lop;
	Expression *_expr2;
      public:
	Logical (Expression *expr1, Op *lop, Expression *expr2);
	~Logical();
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };

    class Bracketed : public Expression
      {
      private:
	Expression *_expr;
      public:
	Bracketed (Expression *expr);
	~Bracketed();
	virtual Query * build_query(Tk::Db::Database *database, const Tk::String &domain) const;
      };
    
    
  }; /* namespace Query */
}; /* namespace Tk */

#endif /* __TK_EXPRESSION_H__ */
