/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <unistd.h>

#include <iostream>
#include <cmath>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "libtkcommon/tkintl.h"
#include "libtkcommon/maths.h"
#include "libtkcommon/real_time.h"
#include "randomness.h"
#include "retriever_registry.h"
#include "random_types.h"

using namespace Tk;
using namespace Random;

#define MAX_BUFFER_SIZE 16384

Randomness:: Randomness(const String &retriever_type, bool deterministic, unsigned buffer_size, bool record_use)
  : _run_state(Stopped),
    _retriever_type(retriever_type),
    _retrieving(0),
    _retrieved(0),
    _buffer_size(buffer_size > MAX_BUFFER_SIZE ? MAX_BUFFER_SIZE : buffer_size),
    _bit_index(0),
    _deterministic(deterministic),
    _needed(buffer_size),
    _used_bits(record_use ? new Bits : 0)
{
  if(!Glib::thread_supported())
    Glib::thread_init();

  unsigned preferred_block_size = RetrieverRegistry::get_preferred_block_size(_retriever_type);
  
  if (deterministic && buffer_size < MAX_BUFFER_SIZE)
    _threshold = 0;
  else if (preferred_block_size >= _buffer_size)
    _threshold = _buffer_size / 2;
  else
    _threshold = _buffer_size - preferred_block_size;
    
  /*
  WDEBUG_OUT << L"Randomness"
	     << L" deterministic: " << deterministic
	     << L"; buffer size parameter: " << buffer_size
	     << L"; actual buffer size: " << _buffer_size
	     << L"; threshold: " << _threshold
	     << L"; preferred block size: " << preferred_block_size
	     << std::endl;
  */

  pthread_mutex_init (&_join_lock, NULL);
  pthread_cond_init (&_join_signal, NULL);

  // fill buffer
  Retriever * retriever = RetrieverRegistry::new_retriever(_retriever_type, _buffer_size);
  retriever->start();
  _retrieving = _buffer_size;
  Buffer * buffer = retriever->join();
  delete retriever;
  append_buffer(buffer);
  delete buffer;

  // fill first randomness
  _randomness = _buffer.back();
  //_randomness = 0xFF;
  _buffer.pop_back();
  --_needed;
  
  // kick off join thread
  _run_state = Running;
  
  //  _thread = Glib::Thread::create(sigc::mem_fun(*this, &Retriever::run), true);
  pthread_create(&_join_thread, NULL, low_join_run, this);
}

Randomness:: ~Randomness()
{
  lock();
  _run_state = Stopped;
  unlock();
  
  //WDEBUG_OUT << L"Signalling join thread..." << std::endl;
  pthread_cond_signal(&_join_signal);

  //WDEBUG_OUT << L"...signalled join thread, joining join thread..." << std::endl;
  int err = pthread_join(_join_thread, NULL);
  if (err)
    { 
      std::wcerr  << __FUNCTION__
		  << L": could not join join thread: ";
      std::cerr   << strerror (err)
		  << std::endl;
      throw UnixError (_("could not join to join thread"), err);
    }
  //WDEBUG_OUT << L"...joined join thread" << std::endl;

  /*  WDEBUG_OUT << "decision retries: "
	     << _decision_tree.get_retries()
	     << " ("
	     << (_count * (_decision_tree.expected_tries() - 1.0))
	     << " expected)"
	     << std::endl;*/
}

void
Randomness:: lock()
{
  //WDEBUG_OUT << "locking" << std::endl;
  pthread_mutex_lock(&_join_lock);
  //WDEBUG_OUT << "locked" << std::endl;
}

void
Randomness:: unlock()
{
  pthread_mutex_unlock(&_join_lock);
  //WDEBUG_OUT << "unlocked" << std::endl;
}

void *
Randomness:: low_join_run(void * data)
{
  Randomness * randomness = static_cast<Randomness *>(data);
  return randomness->join_run();
}

void
Randomness:: append_buffer(const Buffer * buffer)
{
  lock();
  _buffer.insert<Buffer::const_iterator>(_buffer.begin(), buffer->begin(), buffer->end());
  _retrieving -= buffer->size();
  _retrieved += buffer->size();
  unlock();
}

void *
Randomness:: join_run()
{
  Retriever * retriever;
  Buffer * buffer;

  thread_set_real_time(30);

  lock();
  while (_run_state != Stopped)
    {
      if (_retrievers.empty())
	{
	  //WDEBUG_OUT << L"Waiting..." << std::endl;
	  pthread_cond_wait(&_join_signal, &_join_lock);
	  //WDEBUG_OUT << L"...waited" << std::endl;
	}

      if (_retrievers.empty())
	break;

      retriever = _retrievers.front();
      _retrievers.pop_front();
      unlock();
      
      buffer = retriever->join();
      delete retriever;
      append_buffer(buffer);
      delete buffer;
      
      lock();
    }
  unlock();

  //WDEBUG_OUT << L"Join thread finished" << std::endl;
  return NULL;
}

void
Randomness:: retrieve(unsigned count)
{
  Retriever * retriever = RetrieverRegistry::new_retriever(_retriever_type, count);
  retriever->start();

  lock();
  _retrieving += count;
  _retrievers.push_back(retriever);
  unlock();

  pthread_cond_signal(&_join_signal);
}

unsigned
Randomness:: retrieved()
{
  unsigned retrieved;
  lock();
  retrieved = _retrieved;
  unlock();
  return retrieved;
}

Randomness::RunState
Randomness:: run_state()
{
  RunState state;
  lock();
  state = _run_state;
  unlock();
  return state;
}

void
Randomness:: check_threshold()
{
  if (_threshold == 0)
    return;

  unsigned buffered;
  unsigned retrieving;
  
  lock();
  retrieving = _retrieving;
  buffered = _buffer.size();
  unlock();
  
  
  /*
  WDEBUG_OUT << L"buffered: " << buffered
  	     << L"; currently retrieving: " << retrieving
  	     << L"; threshold: " << _threshold
  	     << std::endl;
  */
  
  if (buffered + retrieving <= _threshold)
    {
      //WDEBUG_OUT << L"getting more randomness (" << _buffer_size - retrieving - buffered << L")" << std::endl;
      retrieve(_buffer_size - retrieving - buffered);
      usleep(10);
    }
}

bool
Randomness:: get_bit()
{
  bool bit;

  bit = _randomness & (1 << _bit_index);

  if (_used_bits)
    _used_bits->push_back(bit);

  ++_bit_index;
  
  if (_bit_index == BITS_PER_CHAR && (!_deterministic || _needed > 0))
    {
      _bit_index = 0;
      
      lock();
      //WDEBUG_OUT << "buffer size: " << _buffer.size() << std::endl;
      if (_buffer.empty())
	{
	  unlock();
	  throw std::underflow_error("out of randomness");
	}
      
      _randomness = _buffer.back();
      //WDEBUG_OUT << L"Randomness: `" << static_cast<int>(_randomness) << "'" << std::endl;
      
      _buffer.pop_back();
      unlock();

      --_needed;
      
      if (_threshold != 0 || retrieved() < _buffer_size)
	check_threshold();
    }
  
  return bit;
}

const Bits&
Randomness:: get_used_bits() const
{
  const static Bits null_bits;
  
  if (!_used_bits)
    return null_bits;
  
  return *_used_bits;
}



unsigned long
Randomness:: get_number(unsigned long upper, const String &retriever_type)
{
  static const unsigned short BITS_PER_UL( std::numeric_limits<unsigned long>::digits );
  unsigned long bits = 0;

  Randomness * randomness = new Randomness(retriever_type, true, BITS_PER_UL / BITS_PER_CHAR, false);
  
  
  for (unsigned short i = 0; i < BITS_PER_UL; ++i)
    {
      if (randomness->get_bit())
	bits |= (static_cast<unsigned long>(1) << i);
    }

  delete randomness;

  
  static const unsigned long max( std::numeric_limits<unsigned long>::max() );
  static const long double max_divisor( static_cast<long double>(max) + static_cast<long double>(1.0) );
  long double bits_ratio( static_cast<long double>(bits) / max_divisor );
  long double d_value( static_cast<long double>(upper) * bits_ratio );
  return static_cast<unsigned long>(d_value);
}


/* EOF */
